from wsgiref.headers import Headers
from http.cookies import SimpleCookie
from ario.status import moved_temporarily, moved_permanently
from datetime import datetime, timedelta


class Response(Headers):
    def __init__(self, start_response):
        super().__init__()
        self.__response_encoding = "utf-8"
        self.__cookies = SimpleCookie()
        self.start_response = start_response
        self['Status'] = '200 OK'
        self.is_chunked = False

    @property
    def response_encoding(self):
        return self.__response_encoding

    @response_encoding.setter
    def response_encoding(self, v):
        self.__response_encoding = v

    @property
    def status(self):
        return self['Status']

    @status.setter
    def status(self, v):
        if v is None:
            return
        self['Status'] = v.__str__()

    @response_encoding.setter
    def response_encoding(self, v):
        self.__response_encoding = v

    @property
    def content_type(self):
        content_type = self.get('Content-Type')
        if content_type:
            return content_type.split(';')[0]
        return None

    @content_type.setter
    def content_type(self, v):
        if v is None:
            del self['Content-Type']
        else:
            self['Content-Type'] = '%s; charset=%s' % (v, self.response_encoding)

    @property
    def location(self):
        location = self.get('Location')
        if location:
            return location
        return None

    @location.setter
    def location(self, v):
        if v is None:
            del self['Location']
        else:
            self['Location'] = v

    def perm_redirect(self, location):
        self.location = location
        self.status = moved_permanently()

    def temp_redirect(self, location):
        self.location = location
        self.status = moved_temporarily()

    def cookie(self, name, value, options=None):
        self.__cookies[name] = value
        if options:
            for k, v in options.items():
                self.__cookies[name][k] = v
        return self

    def remove_cookie(self, name):
        self.__cookies[name] = ""
        print(datetime.utcnow())
        expires = datetime.utcnow()
        expires = expires.strftime("%a, %d %b %Y %H:%M:%S GMT")
        self.__cookies[name]['Expires'] = expires

    def encode_response(self, buff):
        if isinstance(buff, bytes):
            return buff
        if self.response_encoding:
            return buff.encode(self.__response_encoding)
        return buff

    def start(self):
        cookie = self.__cookies.output()
        if cookie:
            for line in cookie.split('\r\n'):
                self.add_header(*line.split(': '))
        self.start_response(self.status, self.items())



import os
import mimetypes
from ario.exceptions import BadRequestError

SIZE = 1024 * 5

def serve_static(path, response):
    if not os.path.exists(path):
        raise BadRequestError()
    content_length  = os.path.getsize(path)
    content_type    = mimetypes.guess_type(path)[0]
    response.content_type = content_type
    response.content_length = content_length
    response.is_chunked = True
    def chunk_return():
        with open(path, 'rb') as file:
            while True:
                part = file.read(SIZE)
                if not part:
                    break
                yield part
    return chunk_return



from functools import partial


class StatusCode:
    def __init__(self, code, text):
        self.__code = code
        self.__text = text

    def __str__(self):
        return f"{self.__code} {self.__text}"

    def __repr__(self):
        return f"{self.__code} {self.__text}"


status = StatusCode

ok = partial(status, 200, "OK")

created = partial(status, 201, "CREATED")

accepted = partial(status, 202, "ACCEPTED")

non_authoritative_information = partial(status, 203, "NON AUTHORITATIVE INFORMATION")

no_content = partial(status, 204, "No Content")

reset_content = partial(status, 205, "Reset Content")

partial_content = partial(status, 206, "Partial Content")

multi_status = partial(status, 207, "Multi Status")

multiple_choices = partial(status, 300, "Multiple Choices")

moved_permanently = partial(status, 301, "Moved Permanently")

moved_temporarily = partial(status, 302, "Moved Temporarily")

see_other = partial(status, 303, "See Other")

not_modified = partial(status, 304, "Not Modified")

use_proxy = partial(status, 305, "Use Proxy")

temporary_redirect = partial(status, 307, "Temporary Redirect")

permanent_redirect = partial(status, 308, "Permanent Redirect")

bad_request = partial(status, 400, 'Bad Request')

unauthorized = partial(status, 401, 'Unauthorized')

forbidden = partial(status, 403, 'Forbidden')

not_found = partial(status, 404, 'Not Found')

method_not_allowed = partial(status, 405, 'Method Not Allowed')

conflict = partial(status, 409, 'Conflict')

gone = partial(status, 410, 'Gone')

precondition_failed = partial(status, 412, 'Precondition Failed')

internal_server_error = partial(status, 500, 'Internal Server Error')

bad_gateway = partial(status, 502, 'Bad Gateway')



from http.cookies import SimpleCookie
import ujson 
from urllib.parse import parse_qs
from ario.exceptions import BadRequestError
import wsgiref.util as wsgiutil
import cgi

class Lazy:
    __slots__ = ('f',)

    def __init__(self, f):
        self.f = f

    def __get__(self, obj, t=None):
        f = self.f
        if obj is None:
            return f
        val = f(obj)
        setattr(obj, f.__name__, val)
        return val


class Request:
    def __init__(self, environ):
        self.environ = environ

    @Lazy
    def content_length(self):
        length = self.environ.get("CONTENT_LENGTH")
        if length:
            return int(length.strip())
        return None

    @Lazy
    def content_type(self):
        content_type = self.environ.get("CONTENT_TYPE")
        if content_type:
            return content_type.split(';')[0]
        return None

    @Lazy
    def method(self):
        return self.environ['REQUEST_METHOD'].lower()

    @Lazy
    def path(self):
        return self.environ['PATH_INFO'].lower()

    @Lazy
    def cookies(self):
        cookie = SimpleCookie()
        if 'HTTP_COOKIE' in self.environ:
            cookie.load(self.environ['HTTP_COOKIE'])
        return cookie

    @Lazy
    def body(self):
        if self.content_length is None:
            raise Exception('Content-Length required')
        if self.content_type == "application/json":
            try:
                body = self.environ['wsgi.input'].read(self.content_length)
                body = ujson.decode(body)
            except ValueError as e:
                raise BadRequestError
        else:
            try:
                body = cgi.FieldStorage(
                    fp=self.environ['wsgi.input'],
                    environ=self.environ,
                    keep_blank_values=True
                )
            except (TypeError, ValueError):
                raise BadRequestError
        return body

    @Lazy
    def query(self):
        if 'QUERY_STRING' in self.environ:
            return {k: v[0] if len(v) == 1 else v for k, v in parse_qs(
                self.environ['QUERY_STRING'],
                keep_blank_values=True,
                strict_parsing=False
            ).items()}
        return {}


    @Lazy
    def URI(self):
        return wsgiutil.request_uri(self.environ, include_query=True)



from ario.status import bad_request, unauthorized, forbidden, not_found

class Error(Exception):
    handler = None
    def __init__(self, message=None, status=None):
        self.status = status
        super().__init__(message)


class BadRequestError(Error):
    def __init__(self, message="Bad Request", status=bad_request()):
        self.status = status
        super().__init__(message, status)


class UnauthorizedError(Error):
    def __init__(self, message="Unauthorized", status=unauthorized()):
        self.status = status
        super().__init__(message, status)


class ForbidenError(Error):
    def __init__(self, message="Forbiden", status=forbidden()):
        self.status = status
        super().__init__(message, status)


class NotFoundError(Error):
    def __init__(self, message="Not Found", status=not_found()):
        self.status = status
        super().__init__(message, status)





from dataclasses import dataclass
from typing import List
from ario.request import Request
from ario.response import Response
from ario.status import bad_request
from inspect import signature
from ario.exceptions import Error
import traceback
import json


@dataclass
class Endpoint:
    method: List[str]
    route: str

    def __call__(self):
        print(f"{self.route} called with {self.method}")
        print('ROUTE: ', self.route)

    def __eq__(self, other):
        if (self.method == other.method
                and self.route == other.route):
            return True
        else:
            return False


class RouteNode:
    def __init__(self, method, path, handler=None, child=[]):
        self.method = method
        self.path = path
        self.handler = handler
        self.child = child

    def add_node(self, route, method, handler):
        method = [m.lower() for m in method]
        tokens = RouteNode.__tokenize_route(route)
        if route == "/":
            self.method = method
            self.handler = handler
            return
        routes = self.child
        for i in range(len(tokens)):
            if len(routes) == 0:
                if len(tokens) - 1 == i:
                    node = RouteNode(method, tokens[i], handler, [])
                    if tokens[i][0] == "$":
                        routes.append(node)
                        routes = node.child
                    elif tokens[i][0] == "*":
                        routes.append(node)
                        routes = node.child
                    else:
                        routes.insert(0, node)
                        routes = node.child
                else:
                    node = RouteNode([], tokens[i], [], [])
                    routes.append(node)
                    routes = node.child
                continue
            for (j, r) in enumerate(routes):
                if r.path != tokens[i]:
                    if j == len(routes) - 1:
                        if len(tokens) - 1 == i:
                            node = RouteNode(method, tokens[i], handler, [])
                            if tokens[i][0] == "$":
                                routes.append(node)
                                routes = node.child
                            elif tokens[i][0] == "*":
                                routes.append(node)
                                routes = node.child
                            else:
                                routes.insert(0, node)
                                routes = node.child
                        else:
                            node = RouteNode([], tokens[i], [], [])
                            routes.append(node)
                            routes = node.child
                    continue
                if len(tokens) - 1 == i:
                    r.method = method
                    r.handdler = handler
                else:
                    routes = r.child
                break

    def add_default_node(self, handler):
        self.default = handler

    def find_node(self, route):
        tokens = RouteNode.__tokenize_route(route)
        if route == self.path:
            return (self.handler, self.method, None)
        routes = self.child
        args = None
        for i in range(len(tokens)):
            match_flag = False
            for (j, r) in enumerate(routes):
                if r.path != tokens[i]:
                    if len(tokens) - 1 == i and r.path[0] == '$':
                        return (r.handler, r.method, tokens[i])
                    elif r.path[0] == '*':
                        return (r.handler, r.method, tokens[i:])
                    continue
                match_flag = True
                if len(tokens) - 1 == i:
                    return (r.handler, r.method, None)
                else:
                    routes = r.child
                break
            if not match_flag:
                return None, None, None

    @staticmethod
    def __tokenize_route(route):
        route = route.replace("/", "", 1)
        tokens = route.split("/")
        return tokens

    def __repr__(self):
        return f"path: {self.path}, methods: {self.method}, \n child: {self.child} \n"


class RouterController:
    __instance = None

    @staticmethod
    def get_instance():
        if RouterController.__instance is None:
            RouterController()
        return RouterController.__instance

    def __init__(self, debug=False, langs=[]):
        if RouterController.__instance is None:
            self.__langs = langs
            self.routes = RouteNode([], "/", None)
            self.debug = debug
            RouterController.__instance = self
        else:
            raise Exception("Controller already instantiated")


    def __find_language(self, path):
        path = path.replace("/", "", 1)
        tokens = path.split("/")
        if tokens[0] in self.__langs:
            return tokens[0]
        return None


    def __call__(self, environ, start_response):
        req = Request(environ)
        resp = Response(start_response)
        try:
            method = req.method
            path = req.path

            # in case of /{lang}/other the {lang} is removed and set as an
            # attribute in request and /other is processed as the path
            lang = self.__find_language(path)
            req.lang = lang
            if lang:
                if path == '/' + lang:
                    path = '/'
                else:
                    path = path[len(lang) + 1:]

            handler, methods, arg = self.routes.find_node(path)
            if methods is None or handler is None:
                ret = self.routes.default(req, resp)
            elif method in methods:
                func = getattr(handler, method)
                if arg and len(signature(func).parameters) == 3:
                    ret = func(req, resp, arg)
                else:
                    ret = func(req, resp)
            else:
                ret = self.routes.default(req, resp)
            resp.start()
            if resp.is_chunked:
                return ret()
            ret = resp.encode_response(ret)
            return iter([ret])
        except Error as ex:
            resp.content_type = 'text/html'
            if ex.handler is not None:
                ret = ex.handler()
                resp.start()
            else:
                ret = str(ex.status)
                ret = resp.encode_response(ret)
                resp.start()
            return iter([ret])
        except Exception as ex:
            if self.debug:
                tb = traceback.format_exc()
                tb = resp.encode_response(tb)
                resp.start()
                return iter([tb])
            else:
                ret = self.routes.default(req, resp)
                resp.start()
                return iter([ret])

    def route(self, method=[], route=None):
        def wrapper(handler):
            self.routes.add_node(route, method, handler)

        return wrapper

    def default(self):
        def wrapper(handler):
            self.routes.add_default_node(handler)

        return wrapper



from werkzeug.wrappers import Request, Response


class Application:
    def __init__(self, controller=None):
        if controller is not None:
            self.controller = controller

    def __call__(self, environ, start_response):
        response = self.controller(environ, start_response)
        return response




from functools import wraps
from ario.status import moved_temporarily, moved_permanently
import ujson


def json(handler):
    @wraps(handler)
    def wrapper(request, response, *args):
        response.content_type = 'application/json'
        response.response_encoding = 'utf-8'
        body = handler(request, response, *args)
        body = ujson.dumps(body)
        body = response.encode_response(body)
        return body
    return wrapper


def html(handler):
    @wraps(handler)
    def wrapper(request, response, *args):
        response.content_type = 'text/html'
        response.response_encoding = 'utf-8'
        body = handler(request, response, *args)
        body = response.encode_response(body)
        return body
    return wrapper


def redirect(url, perm=False):
    def decorator(handler):
        @wraps(handler)
        def wrapper(request, response, *args):
            ret = None
            if not perm:
                response.status = moved_temporarily()
                ret = b'302 Moved Temporarily' 
            else:
                response.status = moved_permanently()
                ret = b'301 Moved Permanently' 
            response.location = url
            response.start()
            return ret
        return wrapper
    return decorator



from os.path import dirname, abspath
from jinja2 import Environment, FileSystemLoader, select_autoescape
from functools import wraps

__env = None


def setup_jinja(path):
    global __env
    __env = Environment(loader=FileSystemLoader(path),
                        autoescape=select_autoescape(['html', 'xml']))


def jinja(path):
    def render(func):
        template = __env.get_template(path)

        def handler(request, response, *args):
            response.content_type = 'text/html'
            response.response_encoding = 'utf-8'
            params = func(request, response, *args)
            if not response.location:
                body = template.render(**params)
                body = response.encode_response(body)
            else:
                body = params
            return body

        return handler

    return render



from .application import Application
from .router import Endpoint, RouterController
from .request import Request
from .content import json, html, redirect
from .jinja import jinja, setup_jinja
from .status import *



from setuptools import setup, find_packages
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

with open('requirements.txt') as f:
    required = f.read().splitlines()

setup(
    name="ario",
    version="0.5.0",
    packages=find_packages(),
    author="Wish Team",
    description="A Flask-Based Package for Web Development",
    install_requires=required,
    url="https://github.com/wish-team/ario",
    long_description=long_description,
    long_description_content_type='text/markdown'
)
